#!/usr/bin/env python3

# Copyright (c) 2024 The YAC Authors
#
# SPDX-License-Identifier: BSD-3-Clause

from yac import YAC, Field, TimeUnit, Reduction, InterpolationStack, Points, \
    Action
from threading import Thread
import queue
import json
from isodate import parse_datetime
import numpy as np
import logging

_logger = logging.getLogger('coyote')
_logger.addHandler(logging.NullHandler())


class Coyote:
    """
    COmmon Yac Output Toolkit Environment
    """
    def __init__(self,
                 source_comp : str,
                 source_grid : str,
                 name=None,
                 comm=None,
                 ):
        self.yac = YAC(comm=comm)
        self.source_comp = source_comp
        self.source_grid = source_grid
        self.comp_name = name or f"coyote_{source_comp}_{source_grid}"
        self.comp = self.yac.def_comp(self.comp_name)
        self.target_fields = {}
        self.source_fields = {}
        self.frac_masks = {}
        self.frac_mask_buffers = {}
        _logger.info("Coyote initialized")
        self.yac.sync_def()

    def add_fields(self,
                   variables : list[str],
                   timestep : str,
                   points : Points,
                   interpolation_stack : list,
                   reduction=Reduction.TIME_NONE,
                   frac_mask=None,
                   **coupling_args
                   ):
        if type(interpolation_stack) is list:
            interpolation_stack = InterpolationStack.from_string_json(json.dumps(interpolation_stack))
        assert type(interpolation_stack) is InterpolationStack
        for v in variables:
            collection_size = self.yac.get_field_collection_size(self.source_comp, self.source_grid, v)
            tfield = Field.create(v+"_target", self.comp, points,
                                  collection_size,
                                  timestep, TimeUnit.ISO_FORMAT)
            self.target_fields[v] = tfield
            self.source_fields[v] = Field.create(v, self.comp, points,
                                                 collection_size,
                                                 timestep, TimeUnit.ISO_FORMAT)
            if type(frac_mask) is str:
                self.frac_mask_buffers[frac_mask] = np.empty([points.size])
                self.frac_masks[v] = self.frac_mask_buffers[frac_mask]
            else:
                self.frac_masks[v] = frac_mask
            self.yac.def_couple(self.source_comp, self.source_grid, v,
                                tfield.component_name, tfield.grid_name, tfield.name,
                                timestep, TimeUnit.ISO_FORMAT,
                                reduction, interpolation_stack,
                                **coupling_args)
            _logger.info("Added fields %s", variables)

    def start(self):
        self.yac.enddef()
        q = queue.Queue()
        pass_through_q = queue.Queue()

        def _work():
            while any(f.action != Action.OUT_OF_BOUND for f in self.target_fields.values()):
                datetime = min(parse_datetime(f.datetime) for f in self.target_fields.values())
                _logger.info("Get fields for datetime %s", datetime)
                for name, f in self.target_fields.items():
                    if parse_datetime(f.datetime) == datetime:
                        _logger.info("Get %s", name)
                        data, info = f.get()
                        q.put((name, data,
                               {"action_info": info,
                                "datetime": datetime}))
                        if name in self.frac_mask_buffers.keys():
                            self.frac_mask_buffers[name][:] = data
                        pass_through_q.put((name, data, self.frac_masks[name]))
            _logger.info("Timeloop ended")
            q.put(None)
            q.join()
            pass_through_q.put(None)
            pass_through_q.join()
            _logger.info("worker finished")
        t = Thread(target=_work)
        t.start()

        def _work_pass_through():
            while (elem := pass_through_q.get()) is not None:
                name, data, frac_mask = elem
                _logger.info("Put %s", name)
                self.source_fields[name].put(data, frac_mask=frac_mask)
                pass_through_q.task_done()
            pass_through_q.task_done()
            _logger.info("pass through worker finished")
        pass_through_t = Thread(target=_work_pass_through)
        pass_through_t.start()

        while (elem := q.get()) is not None:
            yield elem
            q.task_done()
        q.task_done()

        _logger.info("waiting for workers")
        t.join()
        pass_through_t.join()
        _logger.info("Done!")
